from __future__ import annotations

import asyncio
from typing import Any

from .context import current_app


def render_template(template_name: str, **context: Any) -> str:
    environment = current_app.get_template_environment()
    template = environment.get_template(template_name)
    return template.render(**context)


async def render_template_async(template_name: str, **context: Any) -> str:
    environment = current_app.get_template_environment()

    def _render() -> str:
        template = environment.get_template(template_name)
        return template.render(**context)

    return await asyncio.to_thread(_render)


def url_for(endpoint: str, **values: Any) -> str:
    return current_app.url_for(endpoint, **values)



from __future__ import annotations

from contextvars import ContextVar, Token
from types import SimpleNamespace
from typing import Any, Callable

_request_var: ContextVar[Any] = ContextVar("exile_request")
_current_app_var: ContextVar[Any] = ContextVar("exile_current_app")
_g_var: ContextVar[SimpleNamespace] = ContextVar("exile_g")


class LocalProxy:
    __slots__ = ("_getter", "_name")

    def __init__(self, getter: Callable[[], Any], name: str) -> None:
        object.__setattr__(self, "_getter", getter)
        object.__setattr__(self, "_name", name)

    def _get_current_object(self) -> Any:
        return object.__getattribute__(self, "_getter")()

    def __getattr__(self, item: str) -> Any:
        return getattr(self._get_current_object(), item)

    def __setattr__(self, key: str, value: Any) -> None:
        setattr(self._get_current_object(), key, value)

    def __delattr__(self, item: str) -> None:
        delattr(self._get_current_object(), item)

    def __repr__(self) -> str:
        try:
            value = self._get_current_object()
        except RuntimeError:
            return f"<LocalProxy {self._name} unbound>"
        return repr(value)

    def __bool__(self) -> bool:
        return bool(self._get_current_object())


def _runtime_error(name: str) -> RuntimeError:
    return RuntimeError(f"Working outside of {name} context.")


def _lookup_request() -> Any:
    try:
        return _request_var.get()
    except LookupError as exc:
        raise _runtime_error("request") from exc


def _lookup_current_app() -> Any:
    try:
        return _current_app_var.get()
    except LookupError as exc:
        raise _runtime_error("app") from exc


def _lookup_g() -> SimpleNamespace:
    try:
        return _g_var.get()
    except LookupError as exc:
        raise _runtime_error("request") from exc


request = LocalProxy(_lookup_request, "request")
current_app = LocalProxy(_lookup_current_app, "current_app")
g = LocalProxy(_lookup_g, "g")


def push_request_context(app: Any, req: Any) -> tuple[Token[Any], Token[Any], Token[SimpleNamespace]]:
    app_token = _current_app_var.set(app)
    request_token = _request_var.set(req)
    g_token = _g_var.set(SimpleNamespace())
    return app_token, request_token, g_token


def pop_request_context(tokens: tuple[Token[Any], Token[Any], Token[SimpleNamespace]]) -> None:
    app_token, request_token, g_token = tokens
    _g_var.reset(g_token)
    _request_var.reset(request_token)
    _current_app_var.reset(app_token)




from __future__ import annotations

import asyncio
import json
import mimetypes
from collections.abc import AsyncIterable, Iterable
from datetime import timezone
from email.utils import formatdate, parsedate_to_datetime
from http.cookies import SimpleCookie
from pathlib import Path
from typing import Any, Literal, Mapping

from .typing import Receive, Scope, Send


class Response:
    def __init__(
        self,
        body: bytes | str = b"",
        status: int = 200,
        headers: Mapping[str, str] | None = None,
        content_type: str | None = "text/plain; charset=utf-8",
    ) -> None:
        if isinstance(body, str):
            self.body = body.encode("utf-8")
        elif isinstance(body, bytes):
            self.body = body
        else:
            raise TypeError("Response body must be str or bytes")

        self.status = int(status)
        self.headers = {str(k): str(v) for k, v in (headers or {}).items()}
        self._extra_headers: list[tuple[str, str]] = []

        has_content_type = any(k.lower() == "content-type" for k in self.headers)
        if content_type is not None and not has_content_type:
            self.headers["Content-Type"] = content_type

    def set_cookie(
        self,
        key: str,
        value: str,
        *,
        max_age: int | None = None,
        expires: str | None = None,
        path: str = "/",
        domain: str | None = None,
        secure: bool = False,
        httponly: bool = False,
        samesite: str | None = "Lax",
    ) -> None:
        cookie = SimpleCookie()
        cookie[key] = value
        morsel = cookie[key]
        morsel["path"] = path
        if max_age is not None:
            morsel["max-age"] = str(max_age)
        if expires is not None:
            morsel["expires"] = expires
        if domain is not None:
            morsel["domain"] = domain
        if secure:
            morsel["secure"] = True
        if httponly:
            morsel["httponly"] = True
        if samesite is not None:
            morsel["samesite"] = samesite
        self._extra_headers.append(("Set-Cookie", morsel.OutputString()))

    async def asgi(self, scope: Scope, receive: Receive, send: Send) -> None:
        _ = scope, receive
        is_head = _scope_method(scope) == "HEAD"
        headers: list[tuple[bytes, bytes]] = []

        has_content_length = False
        for key, value in self.headers.items():
            if key.lower() == "content-length":
                has_content_length = True
            headers.append((key.encode("latin-1"), value.encode("latin-1")))

        if not has_content_length:
            headers.append((b"content-length", str(len(self.body)).encode("latin-1")))

        for key, value in self._extra_headers:
            headers.append((key.encode("latin-1"), value.encode("latin-1")))

        await send(
            {
                "type": "http.response.start",
                "status": self.status,
                "headers": headers,
            }
        )
        await send(
            {
                "type": "http.response.body",
                "body": b"" if is_head else self.body,
                "more_body": False,
            }
        )


class JSONResponse(Response):
    def __init__(
        self,
        content: Any,
        status: int = 200,
        headers: Mapping[str, str] | None = None,
    ) -> None:
        body = json.dumps(content, ensure_ascii=False, separators=(",", ":"))
        super().__init__(
            body=body,
            status=status,
            headers=headers,
            content_type="application/json; charset=utf-8",
        )


class HTMLResponse(Response):
    def __init__(
        self,
        body: bytes | str = "",
        status: int = 200,
        headers: Mapping[str, str] | None = None,
    ) -> None:
        super().__init__(
            body=body,
            status=status,
            headers=headers,
            content_type="text/html; charset=utf-8",
        )


class RedirectResponse(Response):
    def __init__(
        self,
        location: str,
        status: int = 307,
        headers: Mapping[str, str] | None = None,
    ) -> None:
        merged = dict(headers or {})
        merged["Location"] = location
        super().__init__(body=b"", status=status, headers=merged, content_type=None)


class FileResponse(Response):
    def __init__(
        self,
        path: str | Path,
        status: int = 200,
        headers: Mapping[str, str] | None = None,
        content_type: str | None = None,
    ) -> None:
        super().__init__(body=b"", status=status, headers=headers, content_type=content_type)
        self.path = Path(path)

    async def asgi(self, scope: Scope, receive: Receive, send: Send) -> None:
        stat = await asyncio.to_thread(self.path.stat)
        size = stat.st_size

        if _header_lookup(self.headers, "etag") is None:
            self.headers["ETag"] = f'W/"{int(stat.st_mtime)}-{size}"'
        if _header_lookup(self.headers, "last-modified") is None:
            self.headers["Last-Modified"] = formatdate(stat.st_mtime, usegmt=True)
        if _header_lookup(self.headers, "content-length") is None:
            self.headers["Content-Length"] = str(size)
        if _header_lookup(self.headers, "accept-ranges") is None:
            self.headers["Accept-Ranges"] = "bytes"
        if not any(key.lower() == "content-type" for key in self.headers):
            guessed_type, _ = mimetypes.guess_type(str(self.path))
            self.headers["Content-Type"] = guessed_type or "application/octet-stream"

        request_headers = _scope_headers(scope)
        if _is_not_modified(request_headers, etag=_header_lookup(self.headers, "etag"), last_modified=stat.st_mtime):
            self.status = 304
            self.body = b""
            self.headers["Content-Length"] = "0"
            await super().asgi(scope, receive, send)
            return

        range_result = _parse_single_byte_range(request_headers.get("range"), size)
        if range_result == "unsatisfiable":
            self.status = 416
            self.body = b""
            self.headers["Content-Range"] = f"bytes */{size}"
            self.headers["Content-Length"] = "0"
            await super().asgi(scope, receive, send)
            return
        if range_result is not None:
            start, end = range_result
            self.status = 206
            self.headers["Content-Range"] = f"bytes {start}-{end}/{size}"
            self.headers["Content-Length"] = str(end - start + 1)
            if _scope_method(scope) != "HEAD":
                self.body = await asyncio.to_thread(_read_file_range, self.path, start, end - start + 1)
            else:
                self.body = b""
            await super().asgi(scope, receive, send)
            return

        if _scope_method(scope) != "HEAD":
            body = await asyncio.to_thread(self.path.read_bytes)
            self.body = body
        else:
            self.body = b""
        await super().asgi(scope, receive, send)


class StreamingResponse(Response):
    def __init__(
        self,
        content: AsyncIterable[bytes | str] | Iterable[bytes | str],
        status: int = 200,
        headers: Mapping[str, str] | None = None,
        content_type: str | None = "application/octet-stream",
    ) -> None:
        super().__init__(body=b"", status=status, headers=headers, content_type=content_type)
        self.content = content

    async def asgi(self, scope: Scope, receive: Receive, send: Send) -> None:
        _ = scope, receive
        if _scope_method(scope) == "HEAD":
            headers: list[tuple[bytes, bytes]] = []
            for key, value in self.headers.items():
                headers.append((key.encode("latin-1"), value.encode("latin-1")))
            await send(
                {
                    "type": "http.response.start",
                    "status": self.status,
                    "headers": headers,
                }
            )
            await send({"type": "http.response.body", "body": b"", "more_body": False})
            return

        headers: list[tuple[bytes, bytes]] = []
        for key, value in self.headers.items():
            if key.lower() == "content-length":
                continue
            headers.append((key.encode("latin-1"), value.encode("latin-1")))

        await send(
            {
                "type": "http.response.start",
                "status": self.status,
                "headers": headers,
            }
        )

        if hasattr(self.content, "__aiter__"):
            async for chunk in self.content:  # type: ignore[attr-defined]
                payload = chunk.encode("utf-8") if isinstance(chunk, str) else chunk
                await send({"type": "http.response.body", "body": payload, "more_body": True})
        else:
            for chunk in self.content:  # type: ignore[not-an-iterable]
                payload = chunk.encode("utf-8") if isinstance(chunk, str) else chunk
                await send({"type": "http.response.body", "body": payload, "more_body": True})

        await send({"type": "http.response.body", "body": b"", "more_body": False})


def make_response(
    body: Any,
    status: int = 200,
    headers: Mapping[str, str] | None = None,
) -> Response:
    response: Response
    if isinstance(body, Response):
        response = body
        if status != 200:
            response.status = status
    elif isinstance(body, (dict, list)):
        response = JSONResponse(body, status=status, headers=headers)
        return response
    elif isinstance(body, (str, bytes)):
        response = Response(body=body, status=status, headers=headers)
        return response
    else:
        response = Response(body=str(body), status=status, headers=headers)
        return response

    if headers:
        response.headers.update({str(k): str(v) for k, v in headers.items()})
    return response


def jsonify(*args: Any, **kwargs: Any) -> JSONResponse:
    if args and kwargs:
        raise TypeError("jsonify accepts either args or kwargs, not both")
    if len(args) == 1:
        return JSONResponse(args[0])
    if len(args) > 1:
        return JSONResponse(list(args))
    return JSONResponse(kwargs)


def _scope_method(scope: Scope) -> str:
    return str(scope.get("method", "GET")).upper()


def _scope_headers(scope: Scope) -> dict[str, str]:
    raw_headers = scope.get("headers", [])
    parsed: dict[str, str] = {}
    for name, value in raw_headers:
        parsed[name.decode("latin-1").lower()] = value.decode("latin-1")
    return parsed


def _header_lookup(headers: Mapping[str, str], key: str) -> str | None:
    normalized = key.lower()
    for name, value in headers.items():
        if name.lower() == normalized:
            return value
    return None


def _is_not_modified(request_headers: Mapping[str, str], *, etag: str | None, last_modified: float) -> bool:
    if_none_match = request_headers.get("if-none-match")
    if if_none_match:
        candidates = {item.strip() for item in if_none_match.split(",")}
        if "*" in candidates:
            return True
        if etag is not None and etag in candidates:
            return True
        return False

    if_modified_since = request_headers.get("if-modified-since")
    if not if_modified_since:
        return False
    try:
        parsed = parsedate_to_datetime(if_modified_since)
    except (TypeError, ValueError, IndexError):
        return False
    if parsed.tzinfo is None:
        parsed = parsed.replace(tzinfo=timezone.utc)
    return int(last_modified) <= int(parsed.timestamp())


def _parse_single_byte_range(
    range_header: str | None,
    total_size: int,
) -> tuple[int, int] | Literal["unsatisfiable"] | None:
    if not range_header:
        return None
    if not range_header.lower().startswith("bytes="):
        return None

    payload = range_header[len("bytes=") :].strip()
    if not payload or "," in payload or "-" not in payload:
        return None

    start_text, end_text = payload.split("-", 1)
    start_text = start_text.strip()
    end_text = end_text.strip()

    if total_size <= 0:
        return "unsatisfiable"

    if not start_text:
        if not end_text:
            return None
        try:
            suffix_length = int(end_text)
        except ValueError:
            return None
        if suffix_length <= 0:
            return "unsatisfiable"
        if suffix_length >= total_size:
            return (0, total_size - 1)
        return (total_size - suffix_length, total_size - 1)

    try:
        start = int(start_text)
    except ValueError:
        return None
    if start < 0 or start >= total_size:
        return "unsatisfiable"

    if not end_text:
        return (start, total_size - 1)

    try:
        end = int(end_text)
    except ValueError:
        return None
    if end < start:
        return "unsatisfiable"
    return (start, min(end, total_size - 1))


def _read_file_range(path: Path, start: int, length: int) -> bytes:
    with path.open("rb") as file_obj:
        file_obj.seek(start)
        return file_obj.read(length)



from __future__ import annotations

from dataclasses import dataclass
from typing import Any, Callable


@dataclass(slots=True)
class BlueprintRoute:
    rule: str
    methods: list[str] | None
    endpoint: str | None
    handler: Callable[..., Any]


class Blueprint:
    def __init__(
        self,
        name: str,
        import_name: str,
        *,
        url_prefix: str | None = None,
    ) -> None:
        self.name = name
        self.import_name = import_name
        self.url_prefix = url_prefix or ""
        self._routes: list[BlueprintRoute] = []
        self._before_request_funcs: list[Callable[..., Any]] = []
        self._after_request_funcs: list[Callable[..., Any]] = []
        self._teardown_request_funcs: list[Callable[..., Any]] = []
        self._error_handlers: dict[Any, Callable[..., Any]] = {}

    def route(
        self,
        rule: str,
        *,
        methods: list[str] | None = None,
        endpoint: str | None = None,
    ) -> Callable[[Callable[..., Any]], Callable[..., Any]]:
        def decorator(func: Callable[..., Any]) -> Callable[..., Any]:
            self._routes.append(
                BlueprintRoute(rule=rule, methods=methods, endpoint=endpoint, handler=func)
            )
            return func

        return decorator

    def get(self, rule: str, **kwargs: Any) -> Callable[[Callable[..., Any]], Callable[..., Any]]:
        return self.route(rule, methods=["GET"], **kwargs)

    def post(self, rule: str, **kwargs: Any) -> Callable[[Callable[..., Any]], Callable[..., Any]]:
        return self.route(rule, methods=["POST"], **kwargs)

    def put(self, rule: str, **kwargs: Any) -> Callable[[Callable[..., Any]], Callable[..., Any]]:
        return self.route(rule, methods=["PUT"], **kwargs)

    def patch(self, rule: str, **kwargs: Any) -> Callable[[Callable[..., Any]], Callable[..., Any]]:
        return self.route(rule, methods=["PATCH"], **kwargs)

    def delete(self, rule: str, **kwargs: Any) -> Callable[[Callable[..., Any]], Callable[..., Any]]:
        return self.route(rule, methods=["DELETE"], **kwargs)

    def before_request(self, func: Callable[..., Any]) -> Callable[..., Any]:
        self._before_request_funcs.append(func)
        return func

    def after_request(self, func: Callable[..., Any]) -> Callable[..., Any]:
        self._after_request_funcs.append(func)
        return func

    def teardown_request(self, func: Callable[..., Any]) -> Callable[..., Any]:
        self._teardown_request_funcs.append(func)
        return func

    def errorhandler(self, code_or_exception: Any) -> Callable[[Callable[..., Any]], Callable[..., Any]]:
        def decorator(func: Callable[..., Any]) -> Callable[..., Any]:
            self.register_error_handler(code_or_exception, func)
            return func

        return decorator

    def register_error_handler(self, code_or_exception: Any, func: Callable[..., Any]) -> None:
        self._error_handlers[code_or_exception] = func

    def register(self, app: Any, *, url_prefix: str | None = None) -> None:
        prefix = _join_paths(url_prefix or "", self.url_prefix)
        app._register_blueprint_hooks(
            self.name,
            before_request_funcs=self._before_request_funcs,
            after_request_funcs=self._after_request_funcs,
            teardown_request_funcs=self._teardown_request_funcs,
        )
        for key, func in self._error_handlers.items():
            app._register_blueprint_error_handler(self.name, key, func)

        for route in self._routes:
            rule = _join_paths(prefix, route.rule)
            endpoint = route.endpoint or route.handler.__name__
            app.route(
                rule,
                methods=route.methods,
                endpoint=f"{self.name}.{endpoint}",
            )(route.handler)


def _join_paths(prefix: str, path: str) -> str:
    left = prefix.strip("/")
    right = path.strip("/")
    if left and right:
        return f"/{left}/{right}"
    if left:
        return f"/{left}"
    if right:
        return f"/{right}"
    return "/"



from __future__ import annotations

import json
from collections.abc import Iterator, Mapping
from http.cookies import SimpleCookie
from urllib.parse import parse_qsl

from .exceptions import BadRequest
from .typing import Receive, Scope


class MultiDict(Mapping[str, str]):
    def __init__(self, data: dict[str, list[str]] | None = None) -> None:
        self._data = data or {}

    @classmethod
    def from_query_string(cls, query_string: bytes) -> "MultiDict":
        text = query_string.decode("latin-1")
        parsed: dict[str, list[str]] = {}
        for key, value in parse_qsl(text, keep_blank_values=True):
            parsed.setdefault(key, []).append(value)
        return cls(parsed)

    def __getitem__(self, key: str) -> str:
        values = self._data[key]
        return values[0]

    def __iter__(self) -> Iterator[str]:
        return iter(self._data)

    def __len__(self) -> int:
        return len(self._data)

    def getlist(self, key: str) -> list[str]:
        return list(self._data.get(key, []))

    def to_dict(self, flat: bool = True) -> dict[str, str] | dict[str, list[str]]:
        if flat:
            return {k: values[0] for k, values in self._data.items() if values}
        return {k: list(values) for k, values in self._data.items()}


class Request:
    def __init__(self, scope: Scope, receive: Receive) -> None:
        self.scope = scope
        self._receive = receive
        self._body_cache: bytes | None = None
        self._args_cache: MultiDict | None = None
        self._headers_cache: dict[str, str] | None = None

    @property
    def method(self) -> str:
        return str(self.scope.get("method", "GET")).upper()

    @property
    def path(self) -> str:
        return str(self.scope.get("path", "/"))

    @property
    def query_string(self) -> bytes:
        return bytes(self.scope.get("query_string", b""))

    @property
    def headers(self) -> Mapping[str, str]:
        if self._headers_cache is None:
            raw_headers = self.scope.get("headers", [])
            headers: dict[str, str] = {}
            for name, value in raw_headers:
                headers[name.decode("latin-1").lower()] = value.decode("latin-1")
            self._headers_cache = headers
        return self._headers_cache

    @property
    def cookies(self) -> Mapping[str, str]:
        cookie_header = self.headers.get("cookie", "")
        parsed = SimpleCookie()
        parsed.load(cookie_header)
        return {k: morsel.value for k, morsel in parsed.items()}

    @property
    def client(self) -> tuple[str, int] | None:
        client = self.scope.get("client")
        if not isinstance(client, (tuple, list)) or len(client) != 2:
            return None
        host, port = client
        return str(host), int(port)

    @property
    def args(self) -> MultiDict:
        if self._args_cache is None:
            self._args_cache = MultiDict.from_query_string(self.query_string)
        return self._args_cache

    async def body(self) -> bytes:
        if self._body_cache is not None:
            return self._body_cache

        chunks: list[bytes] = []
        more_body = True
        while more_body:
            message = await self._receive()
            message_type = message.get("type")
            if message_type == "http.disconnect":
                break
            if message_type != "http.request":
                continue
            chunks.append(message.get("body", b""))
            more_body = bool(message.get("more_body", False))

        self._body_cache = b"".join(chunks)
        return self._body_cache

    async def text(self, encoding: str = "utf-8") -> str:
        return (await self.body()).decode(encoding)

    async def json(self) -> object:
        raw = await self.body()
        try:
            return json.loads(raw.decode("utf-8"))
        except (UnicodeDecodeError, json.JSONDecodeError) as exc:
            raise BadRequest(description="Invalid JSON body") from exc

    async def form(self) -> dict[str, str | list[str]]:
        raw = await self.body()
        content_type = self.headers.get("content-type", "")
        if "application/x-www-form-urlencoded" not in content_type:
            return {}
        parsed = MultiDict.from_query_string(raw)
        return parsed.to_dict(flat=False)




from __future__ import annotations

import json
from dataclasses import dataclass
from typing import Any, Mapping
from urllib.parse import urlencode

from .typing import ASGIApp, Message


@dataclass(slots=True)
class TestResponse:
    status_code: int
    headers: dict[str, str]
    body: bytes

    @property
    def text(self) -> str:
        return self.body.decode("utf-8")

    def json(self) -> Any:
        return json.loads(self.text)


class AsyncTestClient:
    def __init__(self, app: ASGIApp) -> None:
        self.app = app

    async def request(
        self,
        method: str,
        path: str,
        *,
        headers: Mapping[str, str] | None = None,
        query_params: Mapping[str, Any] | None = None,
        data: bytes | str | None = None,
        json_data: Any = None,
    ) -> TestResponse:
        payload = b""
        request_headers = {k.lower(): v for k, v in (headers or {}).items()}

        if json_data is not None:
            payload = json.dumps(json_data, ensure_ascii=False).encode("utf-8")
            request_headers.setdefault("content-type", "application/json")
        elif isinstance(data, str):
            payload = data.encode("utf-8")
        elif isinstance(data, bytes):
            payload = data

        query_string = b""
        if query_params:
            query_string = urlencode(query_params, doseq=True).encode("utf-8")

        scope = {
            "type": "http",
            "asgi": {"version": "3.0"},
            "http_version": "1.1",
            "method": method.upper(),
            "path": path,
            "query_string": query_string,
            "headers": [(k.encode("latin-1"), v.encode("latin-1")) for k, v in request_headers.items()],
            "client": ("testclient", 0),
            "server": ("testserver", 80),
            "scheme": "http",
        }

        sent_messages: list[Message] = []
        request_sent = False

        async def receive() -> Message:
            nonlocal request_sent
            if request_sent:
                return {"type": "http.disconnect"}
            request_sent = True
            return {"type": "http.request", "body": payload, "more_body": False}

        async def send(message: Message) -> None:
            sent_messages.append(message)

        await self.app(scope, receive, send)

        status_code = 500
        response_headers: dict[str, str] = {}
        body_chunks: list[bytes] = []

        for message in sent_messages:
            message_type = message.get("type")
            if message_type == "http.response.start":
                status_code = int(message.get("status", 500))
                headers_list = message.get("headers", [])
                response_headers = {
                    k.decode("latin-1").lower(): v.decode("latin-1")
                    for k, v in headers_list
                }
            elif message_type == "http.response.body":
                body_chunks.append(message.get("body", b""))

        return TestResponse(
            status_code=status_code,
            headers=response_headers,
            body=b"".join(body_chunks),
        )

    async def get(self, path: str, **kwargs: Any) -> TestResponse:
        return await self.request("GET", path, **kwargs)

    async def post(self, path: str, **kwargs: Any) -> TestResponse:
        return await self.request("POST", path, **kwargs)

    async def put(self, path: str, **kwargs: Any) -> TestResponse:
        return await self.request("PUT", path, **kwargs)

    async def patch(self, path: str, **kwargs: Any) -> TestResponse:
        return await self.request("PATCH", path, **kwargs)

    async def delete(self, path: str, **kwargs: Any) -> TestResponse:
        return await self.request("DELETE", path, **kwargs)




from __future__ import annotations

import importlib.util
import tempfile
import unittest
from pathlib import Path

from exile import AsyncTestClient, Exile, render_template, render_template_async

HAS_JINJA2 = importlib.util.find_spec("jinja2") is not None


@unittest.skipUnless(HAS_JINJA2, "jinja2 is not installed")
class TemplatingTests(unittest.IsolatedAsyncioTestCase):
    async def test_render_template_in_request_context(self) -> None:
        with tempfile.TemporaryDirectory() as tmpdir:
            root = Path(tmpdir)
            (root / "hello.html").write_text("<h1>{{ title }}</h1>", encoding="utf-8")

            app = Exile(__name__, static_folder=None, template_folder=tmpdir)

            @app.get("/hello")
            def hello():
                return render_template("hello.html", title="Exile")

            client = AsyncTestClient(app)
            response = await client.get("/hello")

            self.assertEqual(response.status_code, 200)
            self.assertEqual(response.text, "<h1>Exile</h1>")

    async def test_template_can_use_url_for(self) -> None:
        with tempfile.TemporaryDirectory() as tmpdir:
            root = Path(tmpdir)
            (root / "link.html").write_text("{{ url_for('user_detail', user_id=9) }}", encoding="utf-8")

            app = Exile(__name__, static_folder=None, template_folder=tmpdir)

            @app.get("/users/<int:user_id>")
            def user_detail(user_id: int):
                return {"user_id": user_id}

            @app.get("/link")
            def link():
                return render_template("link.html")

            client = AsyncTestClient(app)
            response = await client.get("/link")

            self.assertEqual(response.status_code, 200)
            self.assertEqual(response.text, "/users/9")

    async def test_render_template_async_helper(self) -> None:
        with tempfile.TemporaryDirectory() as tmpdir:
            root = Path(tmpdir)
            (root / "hello_async.html").write_text("<p>{{ title }}</p>", encoding="utf-8")

            app = Exile(__name__, static_folder=None, template_folder=tmpdir)

            @app.get("/hello-async")
            async def hello_async():
                return await render_template_async("hello_async.html", title="Async")

            client = AsyncTestClient(app)
            response = await client.get("/hello-async")

            self.assertEqual(response.status_code, 200)
            self.assertEqual(response.text, "<p>Async</p>")


if __name__ == "__main__":
    unittest.main()



from __future__ import annotations

from email.utils import formatdate
from pathlib import Path

from .exceptions import NotFound
from .response import FileResponse


def send_from_directory(
    directory: str | Path,
    filename: str,
    *,
    max_age: int = 3600,
) -> FileResponse:
    base = Path(directory).resolve()
    target = (base / filename).resolve()

    if not _is_relative_to(target, base):
        raise NotFound()
    if not target.exists() or not target.is_file():
        raise NotFound()
    return FileResponse(target, headers=_build_cache_headers(target, max_age=max_age))


def _build_cache_headers(path: Path, *, max_age: int) -> dict[str, str]:
    stat = path.stat()
    normalized_max_age = max(0, int(max_age))
    etag = f'W/"{int(stat.st_mtime)}-{stat.st_size}"'
    return {
        "Cache-Control": f"public, max-age={normalized_max_age}",
        "Last-Modified": formatdate(stat.st_mtime, usegmt=True),
        "ETag": etag,
    }


def _is_relative_to(path: Path, parent: Path) -> bool:
    try:
        path.relative_to(parent)
        return True
    except ValueError:
        return False



from __future__ import annotations

import re
import subprocess
import sys
import unittest
from pathlib import Path

import exile


class ReleaseMetadataTests(unittest.TestCase):
    def test_version_sync_between_pyproject_and_init(self) -> None:
        root = Path(__file__).resolve().parents[1]
        pyproject_text = (root / "pyproject.toml").read_text(encoding="utf-8")
        match = re.search(r'(?m)^version\s*=\s*"([^"]+)"\s*$', pyproject_text)
        self.assertIsNotNone(match, "version not found in pyproject.toml")
        self.assertEqual(exile.__version__, match.group(1))

    def test_release_docs_exist(self) -> None:
        root = Path(__file__).resolve().parents[1]
        self.assertTrue((root / "RELEASE.md").exists())
        self.assertTrue((root / "CHANGELOG.md").exists())
        self.assertTrue((root / "exile" / "py.typed").exists())
        self.assertTrue((root / "scripts" / "smoke_test_wheel.py").exists())
        self.assertTrue((root / "scripts" / "build_dists.py").exists())
        self.assertTrue((root / "scripts" / "check_dist_artifacts.py").exists())
        self.assertTrue((root / "scripts" / "publish_testpypi.sh").exists())
        self.assertTrue((root / ".github" / "release.yml").exists())

    def test_changelog_contains_current_version_heading(self) -> None:
        root = Path(__file__).resolve().parents[1]
        changelog_text = (root / "CHANGELOG.md").read_text(encoding="utf-8")
        pattern = rf"(?m)^##\s+\[{re.escape(exile.__version__)}\]\s+-\s+"
        self.assertRegex(changelog_text, pattern)

    def test_version_has_semver_shape(self) -> None:
        self.assertRegex(exile.__version__, r"^\d+\.\d+\.\d+$")

    def test_bump_version_script_dry_run(self) -> None:
        root = Path(__file__).resolve().parents[1]
        script = root / "scripts" / "bump_version.py"
        result = subprocess.run(
            [sys.executable, str(script), "9.9.9", "--dry-run"],
            cwd=root,
            capture_output=True,
            text=True,
            check=False,
        )
        self.assertEqual(result.returncode, 0, msg=result.stderr)
        self.assertIn("pyproject.toml -> version 9.9.9", result.stdout)

    def test_bump_version_script_rejects_invalid_version(self) -> None:
        root = Path(__file__).resolve().parents[1]
        script = root / "scripts" / "bump_version.py"
        result = subprocess.run(
            [sys.executable, str(script), "1.2", "--dry-run"],
            cwd=root,
            capture_output=True,
            text=True,
            check=False,
        )
        self.assertNotEqual(result.returncode, 0)

    def test_changelog_content_check_script_passes(self) -> None:
        root = Path(__file__).resolve().parents[1]
        script = root / "scripts" / "check_changelog_content.py"
        result = subprocess.run(
            [sys.executable, str(script)],
            cwd=root,
            capture_output=True,
            text=True,
            check=False,
        )
        self.assertEqual(result.returncode, 0, msg=result.stderr or result.stdout)


if __name__ == "__main__":
    unittest.main()



from __future__ import annotations

import asyncio
import tempfile
import unittest
from pathlib import Path

from exile import AsyncTestClient, Blueprint, Exile, NotFound, g, request


class FrameworkTests(unittest.IsolatedAsyncioTestCase):
    async def test_dynamic_routing_and_url_for(self) -> None:
        app = Exile(__name__, static_folder=None)

        @app.get("/users/<int:user_id>")
        async def user_detail(user_id: int):
            return {"user_id": user_id}

        client = AsyncTestClient(app)
        response = await client.get("/users/12")

        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.json(), {"user_id": 12})
        self.assertEqual(app.url_for("user_detail", user_id=12, page=2), "/users/12?page=2")

    async def test_head_request_has_no_body(self) -> None:
        app = Exile(__name__, static_folder=None)

        @app.get("/hello")
        def hello():
            return "hello"

        client = AsyncTestClient(app)
        response = await client.request("HEAD", "/hello")

        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.body, b"")
        self.assertEqual(response.headers.get("content-length"), "5")

    async def test_request_helpers_json_form_and_args(self) -> None:
        app = Exile(__name__, static_folder=None)

        @app.post("/echo")
        async def echo():
            payload = await request.json()
            return {
                "payload": payload,
                "q": request.args["q"],
            }

        client = AsyncTestClient(app)
        response = await client.post("/echo", query_params={"q": "ok"}, json_data={"x": 1})
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.json(), {"payload": {"x": 1}, "q": "ok"})

    async def test_hooks_and_context(self) -> None:
        app = Exile(__name__, static_folder=None)
        events: list[str] = []

        @app.before_request
        def before():
            events.append("before")
            g.note = "hello"

        @app.after_request
        def after(resp):
            events.append("after")
            resp.headers["X-After"] = "1"
            return resp

        @app.teardown_request
        def teardown(exc):
            events.append("teardown")
            if exc is not None:
                events.append("teardown-error")

        @app.get("/ok")
        def ok():
            return {"note": g.note}

        client = AsyncTestClient(app)
        response = await client.get("/ok")
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.json(), {"note": "hello"})
        self.assertEqual(response.headers.get("x-after"), "1")
        self.assertEqual(events, ["before", "after", "teardown"])

    async def test_middleware_and_error_handlers(self) -> None:
        app = Exile(__name__, static_folder=None)

        class ASGIHeaderMiddleware:
            def __init__(self, app, *, name: str, value: str):
                self.app = app
                self.name = name
                self.value = value

            async def __call__(self, scope, receive, send):
                async def wrapped_send(message):
                    if message.get("type") == "http.response.start":
                        headers = list(message.get("headers", []))
                        headers.append((self.name.encode("latin-1"), self.value.encode("latin-1")))
                        message = {**message, "headers": headers}
                    await send(message)

                await self.app(scope, receive, wrapped_send)

        app.add_middleware(ASGIHeaderMiddleware, name="X-ASGI", value="1")

        @app.middleware("http")
        async def mw(req, call_next):
            response = await call_next(req)
            response.headers["X-MW"] = "1"
            return response

        @app.errorhandler(RuntimeError)
        def runtime_handler(exc):
            return {"error": str(exc)}, 500

        @app.get("/boom")
        def boom():
            raise RuntimeError("boom")

        @app.get("/ok")
        def ok():
            return {"ok": True}

        client = AsyncTestClient(app)
        ok_response = await client.get("/ok")
        self.assertEqual(ok_response.status_code, 200)
        self.assertEqual(ok_response.json(), {"ok": True})
        self.assertEqual(ok_response.headers.get("x-mw"), "1")
        self.assertEqual(ok_response.headers.get("x-asgi"), "1")

        response = await client.get("/boom")
        self.assertEqual(response.status_code, 500)
        self.assertEqual(response.json(), {"error": "boom"})
        self.assertEqual(response.headers.get("x-asgi"), "1")

    async def test_debug_mode_returns_traceback_for_unhandled_error(self) -> None:
        app = Exile(__name__, static_folder=None, debug=True)

        @app.get("/boom")
        def boom():
            raise RuntimeError("debug-boom")

        client = AsyncTestClient(app)
        response = await client.get("/boom")

        self.assertEqual(response.status_code, 500)
        self.assertIn("RuntimeError: debug-boom", response.text)

    async def test_production_mode_hides_traceback_for_unhandled_error(self) -> None:
        app = Exile(__name__, static_folder=None, debug=False)

        @app.get("/boom")
        def boom():
            raise RuntimeError("prod-boom")

        client = AsyncTestClient(app)
        response = await client.get("/boom")

        self.assertEqual(response.status_code, 500)
        self.assertEqual(response.text, "Internal Server Error")

    async def test_blueprint_hooks_and_error_handlers(self) -> None:
        app = Exile(__name__, static_folder=None)
        bp = Blueprint("api", __name__, url_prefix="/v1")

        @bp.before_request
        def bp_before():
            g.bp = "ok"

        @bp.after_request
        def bp_after(resp):
            resp.headers["X-BP"] = "1"
            return resp

        @bp.errorhandler(404)
        def bp_not_found(exc):
            return {"bp": "not-found"}, 404

        @bp.get("/items/<int:item_id>")
        def item(item_id: int):
            return {"item_id": item_id, "bp": g.bp}

        @bp.get("/missing")
        def missing():
            raise NotFound()

        app.register_blueprint(bp, url_prefix="/api")
        client = AsyncTestClient(app)

        ok_resp = await client.get("/api/v1/items/7")
        self.assertEqual(ok_resp.status_code, 200)
        self.assertEqual(ok_resp.json(), {"item_id": 7, "bp": "ok"})
        self.assertEqual(ok_resp.headers.get("x-bp"), "1")

        miss_resp = await client.get("/api/v1/missing")
        self.assertEqual(miss_resp.status_code, 404)
        self.assertEqual(miss_resp.json(), {"bp": "not-found"})

    async def test_static_file_serving_and_path_traversal_block(self) -> None:
        with tempfile.TemporaryDirectory() as tmpdir:
            root = Path(tmpdir)
            (root / "file.txt").write_text("hello", encoding="utf-8")
            app = Exile(__name__, static_folder=tmpdir)
            client = AsyncTestClient(app)

            ok_resp = await client.get("/static/file.txt")
            self.assertEqual(ok_resp.status_code, 200)
            self.assertEqual(ok_resp.text, "hello")
            self.assertIn("public, max-age=3600", ok_resp.headers.get("cache-control", ""))
            self.assertIn("etag", ok_resp.headers)
            self.assertIn("last-modified", ok_resp.headers)

            bad_resp = await client.get("/static/../file.txt")
            self.assertEqual(bad_resp.status_code, 404)

    async def test_static_file_uses_custom_cache_max_age(self) -> None:
        with tempfile.TemporaryDirectory() as tmpdir:
            root = Path(tmpdir)
            (root / "asset.js").write_text("console.log('ok')", encoding="utf-8")
            app = Exile(__name__, static_folder=tmpdir)
            app.config["STATIC_MAX_AGE"] = 120
            client = AsyncTestClient(app)

            resp = await client.get("/static/asset.js")
            self.assertEqual(resp.status_code, 200)
            self.assertIn("public, max-age=120", resp.headers.get("cache-control", ""))

    async def test_static_file_conditional_request_returns_304(self) -> None:
        with tempfile.TemporaryDirectory() as tmpdir:
            root = Path(tmpdir)
            (root / "bundle.js").write_text("console.log('cache')", encoding="utf-8")
            app = Exile(__name__, static_folder=tmpdir)
            client = AsyncTestClient(app)

            first = await client.get("/static/bundle.js")
            self.assertEqual(first.status_code, 200)
            etag = first.headers.get("etag")
            self.assertIsNotNone(etag)

            second = await client.get("/static/bundle.js", headers={"if-none-match": str(etag)})
            self.assertEqual(second.status_code, 304)
            self.assertEqual(second.body, b"")

    async def test_static_file_if_modified_since_returns_304(self) -> None:
        with tempfile.TemporaryDirectory() as tmpdir:
            root = Path(tmpdir)
            (root / "bundle.css").write_text("body{color:black;}", encoding="utf-8")
            app = Exile(__name__, static_folder=tmpdir)
            client = AsyncTestClient(app)

            first = await client.get("/static/bundle.css")
            self.assertEqual(first.status_code, 200)
            last_modified = first.headers.get("last-modified")
            self.assertIsNotNone(last_modified)

            second = await client.get(
                "/static/bundle.css",
                headers={"if-modified-since": str(last_modified)},
            )
            self.assertEqual(second.status_code, 304)
            self.assertEqual(second.body, b"")

    async def test_static_file_if_modified_since_stale_date_returns_200(self) -> None:
        with tempfile.TemporaryDirectory() as tmpdir:
            root = Path(tmpdir)
            content = "body{background:white;}"
            (root / "bundle.css").write_text(content, encoding="utf-8")
            app = Exile(__name__, static_folder=tmpdir)
            client = AsyncTestClient(app)

            response = await client.get(
                "/static/bundle.css",
                headers={"if-modified-since": "Wed, 21 Oct 2015 07:28:00 GMT"},
            )
            self.assertEqual(response.status_code, 200)
            self.assertEqual(response.text, content)

    async def test_static_file_if_none_match_miss_returns_200(self) -> None:
        with tempfile.TemporaryDirectory() as tmpdir:
            root = Path(tmpdir)
            content = "console.log('etag-miss')"
            (root / "bundle.js").write_text(content, encoding="utf-8")
            app = Exile(__name__, static_folder=tmpdir)
            client = AsyncTestClient(app)

            response = await client.get("/static/bundle.js", headers={"if-none-match": 'W/"0-0"'})
            self.assertEqual(response.status_code, 200)
            self.assertEqual(response.text, content)

    async def test_static_file_if_none_match_wildcard_returns_304(self) -> None:
        with tempfile.TemporaryDirectory() as tmpdir:
            root = Path(tmpdir)
            (root / "wildcard.js").write_text("console.log('wildcard')", encoding="utf-8")
            app = Exile(__name__, static_folder=tmpdir)
            client = AsyncTestClient(app)

            response = await client.get("/static/wildcard.js", headers={"if-none-match": "*"})
            self.assertEqual(response.status_code, 304)
            self.assertEqual(response.body, b"")

    async def test_static_file_if_none_match_list_match_returns_304(self) -> None:
        with tempfile.TemporaryDirectory() as tmpdir:
            root = Path(tmpdir)
            (root / "list.js").write_text("console.log('list')", encoding="utf-8")
            app = Exile(__name__, static_folder=tmpdir)
            client = AsyncTestClient(app)

            first = await client.get("/static/list.js")
            self.assertEqual(first.status_code, 200)
            etag = first.headers.get("etag")
            self.assertIsNotNone(etag)

            response = await client.get(
                "/static/list.js",
                headers={"if-none-match": f'W/"0-0", {etag}, W/"1-1"'},
            )
            self.assertEqual(response.status_code, 304)
            self.assertEqual(response.body, b"")

    async def test_if_none_match_takes_precedence_over_if_modified_since(self) -> None:
        with tempfile.TemporaryDirectory() as tmpdir:
            root = Path(tmpdir)
            content = "console.log('precedence')"
            (root / "precedence.js").write_text(content, encoding="utf-8")
            app = Exile(__name__, static_folder=tmpdir)
            client = AsyncTestClient(app)

            response = await client.get(
                "/static/precedence.js",
                headers={
                    "if-none-match": 'W/"0-0"',
                    "if-modified-since": "Wed, 21 Oct 2099 07:28:00 GMT",
                },
            )
            self.assertEqual(response.status_code, 200)
            self.assertEqual(response.text, content)

    async def test_static_file_range_returns_206_with_partial_content(self) -> None:
        with tempfile.TemporaryDirectory() as tmpdir:
            root = Path(tmpdir)
            content = "abcdefghij"
            (root / "range.txt").write_text(content, encoding="utf-8")
            app = Exile(__name__, static_folder=tmpdir)
            client = AsyncTestClient(app)

            response = await client.get("/static/range.txt", headers={"range": "bytes=2-5"})
            self.assertEqual(response.status_code, 206)
            self.assertEqual(response.text, "cdef")
            self.assertEqual(response.headers.get("content-range"), "bytes 2-5/10")
            self.assertEqual(response.headers.get("content-length"), "4")
            self.assertEqual(response.headers.get("accept-ranges"), "bytes")

    async def test_static_file_head_range_returns_206_without_body(self) -> None:
        with tempfile.TemporaryDirectory() as tmpdir:
            root = Path(tmpdir)
            content = "abcdefghij"
            (root / "range-head.txt").write_text(content, encoding="utf-8")
            app = Exile(__name__, static_folder=tmpdir)
            client = AsyncTestClient(app)

            response = await client.request("HEAD", "/static/range-head.txt", headers={"range": "bytes=1-3"})
            self.assertEqual(response.status_code, 206)
            self.assertEqual(response.body, b"")
            self.assertEqual(response.headers.get("content-range"), "bytes 1-3/10")
            self.assertEqual(response.headers.get("content-length"), "3")

    async def test_static_file_unsatisfiable_range_returns_416(self) -> None:
        with tempfile.TemporaryDirectory() as tmpdir:
            root = Path(tmpdir)
            (root / "range-416.txt").write_text("abc", encoding="utf-8")
            app = Exile(__name__, static_folder=tmpdir)
            client = AsyncTestClient(app)

            response = await client.get("/static/range-416.txt", headers={"range": "bytes=99-100"})
            self.assertEqual(response.status_code, 416)
            self.assertEqual(response.body, b"")
            self.assertEqual(response.headers.get("content-range"), "bytes */3")

    async def test_static_file_head_returns_headers_without_body(self) -> None:
        with tempfile.TemporaryDirectory() as tmpdir:
            root = Path(tmpdir)
            content = "alert('head')"
            (root / "head.js").write_text(content, encoding="utf-8")
            app = Exile(__name__, static_folder=tmpdir)
            client = AsyncTestClient(app)

            response = await client.request("HEAD", "/static/head.js")
            self.assertEqual(response.status_code, 200)
            self.assertEqual(response.body, b"")
            self.assertEqual(response.headers.get("content-length"), str(len(content.encode("utf-8"))))

    async def test_lifespan_events(self) -> None:
        app = Exile(__name__, static_folder=None)
        events: list[str] = []

        @app.on_startup
        def startup():
            events.append("startup")

        @app.on_shutdown
        async def shutdown():
            events.append("shutdown")

        sent = []
        incoming = iter(
            [
                {"type": "lifespan.startup"},
                {"type": "lifespan.shutdown"},
            ]
        )

        async def receive():
            return next(incoming)

        async def send(message):
            sent.append(message)

        await app({"type": "lifespan"}, receive, send)
        self.assertEqual(
            sent,
            [
                {"type": "lifespan.startup.complete"},
                {"type": "lifespan.shutdown.complete"},
            ],
        )
        self.assertEqual(events, ["startup", "shutdown"])

    def test_context_proxy_outside_request_raises(self) -> None:
        with self.assertRaises(RuntimeError):
            _ = request.path


if __name__ == "__main__":
    unittest.main()



from .app import Exile
from .blueprints import Blueprint
from .context import current_app, g, request
from .exceptions import HTTPException, MethodNotAllowed, NotFound, abort
from .request import Request
from .response import (
    HTMLResponse,
    JSONResponse,
    RedirectResponse,
    Response,
    FileResponse,
    StreamingResponse,
    jsonify,
    make_response,
)
from .staticfiles import send_from_directory
from .testing import AsyncTestClient, TestResponse
from .templating import render_template, render_template_async, url_for

__version__ = "0.1.0"

__all__ = [
    "__version__",
    "Exile",
    "Blueprint",
    "request",
    "current_app",
    "g",
    "Request",
    "Response",
    "FileResponse",
    "StreamingResponse",
    "JSONResponse",
    "HTMLResponse",
    "RedirectResponse",
    "render_template",
    "render_template_async",
    "url_for",
    "send_from_directory",
    "AsyncTestClient",
    "TestResponse",
    "abort",
    "jsonify",
    "make_response",
    "HTTPException",
    "NotFound",
    "MethodNotAllowed",
]



from __future__ import annotations

from collections.abc import Iterable
from typing import NoReturn


class HTTPException(Exception):
    status_code = 500
    default_description = "Internal Server Error"

    def __init__(
        self,
        status_code: int | None = None,
        description: str | None = None,
        headers: dict[str, str] | None = None,
    ) -> None:
        self.status_code = status_code or self.status_code
        self.description = description or self.default_description
        self.headers = headers or {}
        super().__init__(self.description)


class BadRequest(HTTPException):
    status_code = 400
    default_description = "Bad Request"


class NotFound(HTTPException):
    status_code = 404
    default_description = "Not Found"


class MethodNotAllowed(HTTPException):
    status_code = 405
    default_description = "Method Not Allowed"

    def __init__(
        self,
        *,
        allowed_methods: Iterable[str] | None = None,
        description: str | None = None,
    ) -> None:
        methods = sorted({m.upper() for m in (allowed_methods or [])})
        headers = {"Allow": ", ".join(methods)} if methods else {}
        super().__init__(description=description, headers=headers)


class InternalServerError(HTTPException):
    status_code = 500
    default_description = "Internal Server Error"


def abort(status_code: int, description: str | None = None) -> NoReturn:
    raise HTTPException(status_code=status_code, description=description)




from __future__ import annotations

from typing import Any, Awaitable, Callable, TypeAlias

Scope: TypeAlias = dict[str, Any]
Message: TypeAlias = dict[str, Any]
Receive: TypeAlias = Callable[[], Awaitable[Message]]
Send: TypeAlias = Callable[[Message], Awaitable[None]]
ASGIApp: TypeAlias = Callable[[Scope, Receive, Send], Awaitable[None]]




from __future__ import annotations

import re
from collections.abc import Callable
from dataclasses import dataclass
from typing import Any, Pattern
from urllib.parse import urlencode


_DYNAMIC_SEGMENT_RE = re.compile(r"<(?:(?P<converter>[a-zA-Z_][a-zA-Z0-9_]*):)?(?P<name>[a-zA-Z_][a-zA-Z0-9_]*)>")


@dataclass(slots=True)
class CompiledRule:
    rule: str
    pattern: Pattern[str]
    converters: dict[str, Callable[[str], Any]]


def compile_rule(rule: str) -> CompiledRule:
    if not rule.startswith("/"):
        raise ValueError(f"route must start with '/': {rule!r}")

    converters: dict[str, Callable[[str], Any]] = {}
    regex_parts: list[str] = []
    last_index = 0

    for match in _DYNAMIC_SEGMENT_RE.finditer(rule):
        start, end = match.span()
        regex_parts.append(re.escape(rule[last_index:start]))

        converter_name = match.group("converter") or "str"
        param_name = match.group("name")
        param_pattern, converter = _resolve_converter(converter_name)
        regex_parts.append(f"(?P<{param_name}>{param_pattern})")
        converters[param_name] = converter
        last_index = end

    regex_parts.append(re.escape(rule[last_index:]))
    pattern = re.compile(f"^{''.join(regex_parts)}$")
    return CompiledRule(rule=rule, pattern=pattern, converters=converters)


def convert_matched_params(compiled_rule: CompiledRule, values: dict[str, str]) -> dict[str, Any]:
    parsed: dict[str, Any] = {}
    for key, value in values.items():
        converter = compiled_rule.converters.get(key, str)
        parsed[key] = converter(value)
    return parsed


def build_rule_path(rule: str, values: dict[str, Any]) -> str:
    remaining = dict(values)
    chunks: list[str] = []
    last_index = 0

    for match in _DYNAMIC_SEGMENT_RE.finditer(rule):
        start, end = match.span()
        chunks.append(rule[last_index:start])
        param_name = match.group("name")
        if param_name not in remaining:
            raise KeyError(param_name)
        chunks.append(str(remaining.pop(param_name)))
        last_index = end

    chunks.append(rule[last_index:])
    path = "".join(chunks)

    if remaining:
        query = urlencode(remaining, doseq=True)
        if query:
            return f"{path}?{query}"
    return path


def _resolve_converter(name: str) -> tuple[str, Callable[[str], Any]]:
    if name == "str":
        return r"[^/]+", str
    if name == "int":
        return r"\d+", int
    if name == "float":
        return r"\d+(?:\.\d+)?", float
    if name == "path":
        return r".+", str
    raise ValueError(f"unsupported converter: {name}")



from __future__ import annotations

import inspect
import traceback
from dataclasses import dataclass
from pathlib import Path
from typing import Any, Awaitable, Callable, Mapping

from .context import pop_request_context, push_request_context
from .exceptions import HTTPException, InternalServerError, MethodNotAllowed, NotFound
from .request import Request
from .response import Response, make_response
from .routing import CompiledRule, build_rule_path, compile_rule, convert_matched_params
from .typing import Message, Receive, Scope, Send

CallNext = Callable[[Request], Awaitable[Response]]
HTTPMiddlewareFunc = Callable[[Request, CallNext], Any]


@dataclass(slots=True)
class Route:
    rule: str
    methods: set[str]
    endpoint: str
    handler: Callable[..., Any]
    compiled_rule: CompiledRule


class Exile:
    def __init__(
        self,
        import_name: str,
        *,
        debug: bool = False,
        static_folder: str | None = "static",
        template_folder: str | None = "templates",
    ) -> None:
        self.import_name = import_name
        self.debug = debug
        self.static_folder = static_folder
        self.template_folder = template_folder
        self.config: dict[str, Any] = {
            "STATIC_MAX_AGE": 3600,
        }

        self._routes: list[Route] = []
        self._routes_by_endpoint: dict[str, Route] = {}
        self._error_handlers: dict[Any, Callable[..., Any]] = {}
        self._before_request_funcs: list[Callable[..., Any]] = []
        self._after_request_funcs: list[Callable[..., Any]] = []
        self._teardown_request_funcs: list[Callable[..., Any]] = []
        self._blueprint_before_request_funcs: dict[str, list[Callable[..., Any]]] = {}
        self._blueprint_after_request_funcs: dict[str, list[Callable[..., Any]]] = {}
        self._blueprint_teardown_request_funcs: dict[str, list[Callable[..., Any]]] = {}
        self._blueprint_error_handlers: dict[str, dict[Any, Callable[..., Any]]] = {}
        self._http_middlewares: list[HTTPMiddlewareFunc] = []
        self._asgi_middleware_defs: list[tuple[Callable[..., Any], dict[str, Any]]] = []
        self._asgi_app_cache: Callable[[Scope, Receive, Send], Awaitable[None]] | None = None
        self._template_env: Any = None
        self._startup_funcs: list[Callable[..., Any]] = []
        self._shutdown_funcs: list[Callable[..., Any]] = []

        if self.static_folder:
            self._register_default_static_route()

    async def __call__(self, scope: Scope, receive: Receive, send: Send) -> None:
        app = self._build_asgi_app()
        await app(scope, receive, send)

    def route(
        self,
        rule: str,
        *,
        methods: list[str] | None = None,
        endpoint: str | None = None,
    ) -> Callable[[Callable[..., Any]], Callable[..., Any]]:
        allowed_methods = {m.upper() for m in (methods or ["GET"])}
        if "GET" in allowed_methods:
            allowed_methods.add("HEAD")
        compiled_rule = compile_rule(rule)

        def decorator(func: Callable[..., Any]) -> Callable[..., Any]:
            route_obj = Route(
                rule=rule,
                methods=set(allowed_methods),
                endpoint=endpoint or func.__name__,
                handler=func,
                compiled_rule=compiled_rule,
            )
            self._routes.append(route_obj)
            self._routes_by_endpoint[route_obj.endpoint] = route_obj
            return func

        return decorator

    def get(self, rule: str, **kwargs: Any) -> Callable[[Callable[..., Any]], Callable[..., Any]]:
        return self.route(rule, methods=["GET"], **kwargs)

    def post(self, rule: str, **kwargs: Any) -> Callable[[Callable[..., Any]], Callable[..., Any]]:
        return self.route(rule, methods=["POST"], **kwargs)

    def put(self, rule: str, **kwargs: Any) -> Callable[[Callable[..., Any]], Callable[..., Any]]:
        return self.route(rule, methods=["PUT"], **kwargs)

    def patch(self, rule: str, **kwargs: Any) -> Callable[[Callable[..., Any]], Callable[..., Any]]:
        return self.route(rule, methods=["PATCH"], **kwargs)

    def delete(self, rule: str, **kwargs: Any) -> Callable[[Callable[..., Any]], Callable[..., Any]]:
        return self.route(rule, methods=["DELETE"], **kwargs)

    def options(self, rule: str, **kwargs: Any) -> Callable[[Callable[..., Any]], Callable[..., Any]]:
        return self.route(rule, methods=["OPTIONS"], **kwargs)

    def head(self, rule: str, **kwargs: Any) -> Callable[[Callable[..., Any]], Callable[..., Any]]:
        return self.route(rule, methods=["HEAD"], **kwargs)

    def before_request(self, func: Callable[..., Any]) -> Callable[..., Any]:
        self._before_request_funcs.append(func)
        return func

    def after_request(self, func: Callable[..., Any]) -> Callable[..., Any]:
        self._after_request_funcs.append(func)
        return func

    def teardown_request(self, func: Callable[..., Any]) -> Callable[..., Any]:
        self._teardown_request_funcs.append(func)
        return func

    def middleware(self, kind: str = "http") -> Callable[[HTTPMiddlewareFunc], HTTPMiddlewareFunc]:
        normalized = kind.lower()
        if normalized != "http":
            raise ValueError(f"unsupported middleware type: {kind}")

        def decorator(func: HTTPMiddlewareFunc) -> HTTPMiddlewareFunc:
            self._http_middlewares.append(func)
            return func

        return decorator

    def add_middleware(self, middleware_cls: Callable[..., Any], **options: Any) -> None:
        self._asgi_middleware_defs.append((middleware_cls, dict(options)))
        self._asgi_app_cache = None

    def errorhandler(self, code_or_exception: Any) -> Callable[[Callable[..., Any]], Callable[..., Any]]:
        def decorator(func: Callable[..., Any]) -> Callable[..., Any]:
            self.register_error_handler(code_or_exception, func)
            return func

        return decorator

    def register_error_handler(self, code_or_exception: Any, func: Callable[..., Any]) -> None:
        self._error_handlers[code_or_exception] = func

    def register_blueprint(self, blueprint: Any, *, url_prefix: str | None = None) -> None:
        blueprint.register(self, url_prefix=url_prefix)

    def on_startup(self, func: Callable[..., Any]) -> Callable[..., Any]:
        self._startup_funcs.append(func)
        return func

    def on_shutdown(self, func: Callable[..., Any]) -> Callable[..., Any]:
        self._shutdown_funcs.append(func)
        return func

    def _register_blueprint_hooks(
        self,
        blueprint_name: str,
        *,
        before_request_funcs: list[Callable[..., Any]] | None = None,
        after_request_funcs: list[Callable[..., Any]] | None = None,
        teardown_request_funcs: list[Callable[..., Any]] | None = None,
    ) -> None:
        self._blueprint_before_request_funcs.setdefault(blueprint_name, []).extend(before_request_funcs or [])
        self._blueprint_after_request_funcs.setdefault(blueprint_name, []).extend(after_request_funcs or [])
        self._blueprint_teardown_request_funcs.setdefault(blueprint_name, []).extend(teardown_request_funcs or [])

    def _register_blueprint_error_handler(
        self,
        blueprint_name: str,
        code_or_exception: Any,
        func: Callable[..., Any],
    ) -> None:
        handlers = self._blueprint_error_handlers.setdefault(blueprint_name, {})
        handlers[code_or_exception] = func

    def run(
        self,
        host: str = "127.0.0.1",
        port: int = 8000,
        *,
        reload: bool = False,
    ) -> None:
        try:
            import uvicorn
        except ModuleNotFoundError as exc:
            raise RuntimeError("uvicorn is required for app.run()") from exc
        uvicorn.run(self, host=host, port=port, reload=reload)

    def url_for(self, endpoint: str, **values: Any) -> str:
        route = self._routes_by_endpoint.get(endpoint)
        if route is None:
            raise KeyError(f"unknown endpoint: {endpoint}")
        return build_rule_path(route.rule, values)

    def get_template_environment(self) -> Any:
        if self._template_env is not None:
            return self._template_env

        if not self.template_folder:
            raise RuntimeError("template folder is disabled")
        try:
            from jinja2 import Environment, FileSystemLoader
        except ModuleNotFoundError as exc:
            raise RuntimeError("jinja2 is required for template rendering") from exc

        env = Environment(loader=FileSystemLoader(str(self.template_folder_path)), autoescape=True)
        env.globals["url_for"] = self.url_for
        self._template_env = env
        return env

    def _build_asgi_app(self) -> Callable[[Scope, Receive, Send], Awaitable[None]]:
        if self._asgi_app_cache is not None:
            return self._asgi_app_cache

        app: Callable[[Scope, Receive, Send], Awaitable[None]] = self._dispatch_asgi
        for middleware_cls, options in reversed(self._asgi_middleware_defs):
            app = middleware_cls(app, **options)

        self._asgi_app_cache = app
        return app

    async def _dispatch_asgi(self, scope: Scope, receive: Receive, send: Send) -> None:
        scope_type = scope.get("type")
        if scope_type == "http":
            await self._handle_http(scope, receive, send)
            return
        if scope_type == "lifespan":
            await self._handle_lifespan(receive, send)
            return
        if scope_type == "websocket":
            await send({"type": "websocket.close", "code": 1000})
            return

    async def _handle_http(self, scope: Scope, receive: Receive, send: Send) -> None:
        request = Request(scope, receive)
        context_tokens = push_request_context(self, request)
        err: Exception | None = None

        try:
            response = await self._run_http_middlewares(request)
        except Exception as exc:
            err = exc
            response = await self._handle_exception(exc, request=request)
        finally:
            await self._run_teardown_request(err, blueprint_name=self._get_request_blueprint(request))
            pop_request_context(context_tokens)

        await response.asgi(scope, receive, send)

    async def _handle_lifespan(self, receive: Receive, send: Send) -> None:
        while True:
            message: Message = await receive()
            message_type = message.get("type")
            if message_type == "lifespan.startup":
                try:
                    for func in self._startup_funcs:
                        await self._invoke_hook(func)
                except Exception as exc:
                    await send({"type": "lifespan.startup.failed", "message": str(exc)})
                    continue
                await send({"type": "lifespan.startup.complete"})
            elif message_type == "lifespan.shutdown":
                try:
                    for func in self._shutdown_funcs:
                        await self._invoke_hook(func)
                except Exception as exc:
                    await send({"type": "lifespan.shutdown.failed", "message": str(exc)})
                    return
                await send({"type": "lifespan.shutdown.complete"})
                return

    async def _run_http_middlewares(self, request: Request) -> Response:
        next_handler: CallNext = self._dispatch_request
        for middleware in reversed(self._http_middlewares):
            downstream = next_handler

            async def _wrap(req: Request, middleware: HTTPMiddlewareFunc = middleware, downstream: CallNext = downstream) -> Response:
                result = await self._invoke(middleware, req, downstream)
                return self._coerce_response(result)

            next_handler = _wrap
        return await next_handler(request)

    async def _dispatch_request(self, request: Request) -> Response:
        before_result = await self._run_before_request()
        if before_result is not None:
            response = self._coerce_response(before_result)
            return await self._run_after_request(response, blueprint_name=None)

        route, path_params, allowed = self._match_route(path=request.path, method=request.method)
        if route is None:
            if request.method == "OPTIONS" and allowed:
                allow = ", ".join(sorted(allowed | {"OPTIONS"}))
                return make_response("", status=204, headers={"Allow": allow})
            if allowed:
                raise MethodNotAllowed(allowed_methods=allowed)
            raise NotFound()

        blueprint_name = self._extract_blueprint_name(route.endpoint)
        request.scope["_exile.endpoint"] = route.endpoint
        request.scope["_exile.blueprint"] = blueprint_name

        before_blueprint_result = await self._run_blueprint_before_request(blueprint_name)
        if before_blueprint_result is not None:
            response = self._coerce_response(before_blueprint_result)
            return await self._run_after_request(response, blueprint_name=blueprint_name)

        result = await self._invoke(route.handler, **path_params)
        response = self._coerce_response(result)
        return await self._run_after_request(response, blueprint_name=blueprint_name)

    async def _invoke(self, func: Callable[..., Any], *args: Any, **kwargs: Any) -> Any:
        result = func(*args, **kwargs)
        if inspect.isawaitable(result):
            return await result
        return result

    async def _invoke_hook(self, func: Callable[..., Any], *preferred_args: Any) -> Any:
        signature = inspect.signature(func)
        params = list(signature.parameters.values())
        has_var_args = any(param.kind == inspect.Parameter.VAR_POSITIONAL for param in params)
        if has_var_args:
            args = preferred_args
        else:
            positional = [
                param
                for param in params
                if param.kind in (inspect.Parameter.POSITIONAL_ONLY, inspect.Parameter.POSITIONAL_OR_KEYWORD)
            ]
            args = preferred_args[: len(positional)]
        return await self._invoke(func, *args)

    def _match_route(self, *, path: str, method: str) -> tuple[Route | None, dict[str, Any], set[str]]:
        allowed_methods: set[str] = set()
        for route in self._routes:
            match = route.compiled_rule.pattern.fullmatch(path)
            if match is None:
                continue
            if method in route.methods:
                params = convert_matched_params(route.compiled_rule, match.groupdict())
                return route, params, set()
            allowed_methods.update(route.methods)
        return None, {}, allowed_methods

    def _coerce_response(self, result: Any) -> Response:
        if isinstance(result, tuple):
            if len(result) == 2:
                body, status = result
                return make_response(body, status=int(status))
            if len(result) == 3:
                body, status, headers = result
                if not isinstance(headers, Mapping):
                    raise TypeError("response headers must be a mapping")
                return make_response(body, status=int(status), headers=headers)
            raise TypeError("tuple responses must be (body, status) or (body, status, headers)")
        return make_response(result)

    async def _run_before_request(self) -> Any:
        for func in self._before_request_funcs:
            result = await self._invoke_hook(func)
            if result is not None:
                return result
        return None

    async def _run_blueprint_before_request(self, blueprint_name: str | None) -> Any:
        if not blueprint_name:
            return None
        for func in self._blueprint_before_request_funcs.get(blueprint_name, []):
            result = await self._invoke_hook(func)
            if result is not None:
                return result
        return None

    async def _run_after_request(self, response: Response, *, blueprint_name: str | None) -> Response:
        current = response
        if blueprint_name:
            for func in self._blueprint_after_request_funcs.get(blueprint_name, []):
                maybe_response = await self._invoke_hook(func, current)
                if maybe_response is not None:
                    current = self._coerce_response(maybe_response)
        for func in self._after_request_funcs:
            maybe_response = await self._invoke_hook(func, current)
            if maybe_response is not None:
                current = self._coerce_response(maybe_response)
        return current

    async def _run_teardown_request(self, error: Exception | None, *, blueprint_name: str | None) -> None:
        if blueprint_name:
            for func in self._blueprint_teardown_request_funcs.get(blueprint_name, []):
                await self._invoke_hook(func, error)
        for func in self._teardown_request_funcs:
            await self._invoke_hook(func, error)

    async def _handle_exception(self, exc: Exception, *, request: Request | None = None) -> Response:
        blueprint_name = self._get_request_blueprint(request) if request else None
        handler = self._find_error_handler(exc, blueprint_name=blueprint_name)
        if handler is not None:
            return self._coerce_response(await self._invoke_hook(handler, exc))

        if isinstance(exc, HTTPException):
            return make_response(exc.description, status=exc.status_code, headers=exc.headers)

        internal_error = InternalServerError()
        if self.debug:
            stack = traceback.format_exc()
            return make_response(stack, status=500, headers={"Content-Type": "text/plain; charset=utf-8"})
        return make_response(internal_error.description, status=internal_error.status_code)

    def _find_error_handler(self, exc: Exception, *, blueprint_name: str | None = None) -> Callable[..., Any] | None:
        if blueprint_name:
            blueprint_handlers = self._blueprint_error_handlers.get(blueprint_name, {})
            handler = self._lookup_error_handler(blueprint_handlers, exc)
            if handler is not None:
                return handler

        return self._lookup_error_handler(self._error_handlers, exc)

    def _lookup_error_handler(
        self,
        registry: Mapping[Any, Callable[..., Any]],
        exc: Exception,
    ) -> Callable[..., Any] | None:
        if isinstance(exc, HTTPException):
            code_handler = registry.get(exc.status_code)
            if code_handler is not None:
                return code_handler

        for cls in type(exc).__mro__:
            handler = registry.get(cls)
            if handler is not None:
                return handler
        return None

    def _extract_blueprint_name(self, endpoint: str) -> str | None:
        if "." not in endpoint:
            return None
        return endpoint.split(".", 1)[0]

    def _get_request_blueprint(self, request: Request | None) -> str | None:
        if request is None:
            return None
        value = request.scope.get("_exile.blueprint")
        if isinstance(value, str):
            return value
        return None

    def _register_default_static_route(self) -> None:
        async def _serve_static(filename: str) -> Response:
            from .staticfiles import send_from_directory

            configured_max_age = self.config.get("STATIC_MAX_AGE", 3600)
            try:
                max_age = int(configured_max_age)
            except (TypeError, ValueError):
                max_age = 3600

            return send_from_directory(self.static_folder_path, filename, max_age=max_age)

        self.route("/static/<path:filename>", methods=["GET"], endpoint="static")(_serve_static)

    @property
    def static_folder_path(self) -> Path:
        if not self.static_folder:
            raise RuntimeError("static folder is disabled")
        return Path(self.static_folder)

    @property
    def template_folder_path(self) -> Path:
        if not self.template_folder:
            raise RuntimeError("template folder is disabled")
        return Path(self.template_folder)
